@@ -1,3 +1,5 @@
+# -*- coding: utf-8 -*-
+
 #
 # furl - URL manipulation made simple.
 #
@@ -53,16 +55,298 @@ DEFAULT_PORTS = {
 # TODO(grun): Support schemes separated by just ':', not '://' without having
 # an explicit list. There are many such schemes in various URIs.
 COLON_SEPARATED_SCHEMES = [
-    'mailto',
-    'tel',
     'sms',
+    'tel',
+    'mailto',
 ]
 
 
+def lget(l, index, default=None):
+    try:
+        return l[index]
+    except IndexError:
+        return default
+
+
+def attemptstr(o):
+    try:
+        return str(o)
+    except:
+        return o
+
+
+def utf8(o, default=_absent):
+    try:
+        return o.encode('utf8')
+    except:
+        return o if default is _absent else default
+
+
+def callable_attr(obj, attr):
+    return hasattr(obj, attr) and callable(getattr(obj, attr))
+
+
 def non_text_iterable(value):
     b = callable_attr(value, '__iter__') and not isinstance(value, basestring)
     return b
-        
+
+
+# TODO(grun): Support IDNA2008 via the third party idna module. See
+# https://github.com/gruns/furl/issues/73.
+def idna_encode(o):
+    if callable_attr(o, 'encode'):
+        return o.encode('idna').decode('utf8')
+    return o
+
+
+def idna_decode(o):
+    if callable_attr(utf8(o), 'decode'):
+        return utf8(o).decode('idna')
+    return o
+
+
+def is_valid_port(port):
+    port = str(port)
+    if not port.isdigit() or not 0 < int(port) <= 65535:
+        return False
+    return True
+
+
+def static_vars(**kwargs):
+    def decorator(func):
+        for key, value in six.iteritems(kwargs):
+            setattr(func, key, value)
+        return func
+    return decorator
+
+#
+# TODO(grun): Update some of the regex functions below to reflect the fact that
+# the valid encoding of Path segments differs slightly from the valid encoding
+# of Fragment Path segments. Similarly, the valid encodings of Query keys and
+# values differ slightly from the valid encodings of Fragment Query keys and
+# values.
+#
+# For example, '?' and '#' don't need to be encoded in Fragment Path segments
+# but they must be encoded in Path segments. Similarly, '#' doesn't need to be
+# encoded in Fragment Query keys and values, but must be encoded in Query keys
+# and values.
+#
+# Perhaps merge them with URLPath, FragmentPath, URLQuery, and
+# FragmentQuery when those new classes are created (see the TODO
+# currently at the top of the source, 02/03/2012).
+#
+
+# RFC 3986
+#   unreserved  = ALPHA / DIGIT / "-" / "." / "_" / "~"
+#
+#   pct-encoded = "%" HEXDIG HEXDIG
+#
+#   sub-delims  = "!" / "$" / "&" / "'" / "(" / ")"
+#                 / "*" / "+" / "," / ";" / "="
+#
+#   pchar       = unreserved / pct-encoded / sub-delims / ":" / "@"
+#
+#   === Path ===
+#   segment     = *pchar
+#
+#   === Query ===
+#   query       = *( pchar / "/" / "?" )
+#
+#   === Scheme ===
+#   scheme      = ALPHA *( ALPHA / DIGIT / "+" / "-" / "." )
+#
+PERCENT_REGEX = r'\%[a-fA-F\d][a-fA-F\d]'
+INVALID_DOMAIN_CHARS = '!@#$%^&\'\"*()+=:;/'
+
+
+@static_vars(regex=re.compile(
+    r'^([\w%s]|(%s))*$' % (re.escape('-.~:@!$&\'()*+,;='), PERCENT_REGEX)))
+def is_valid_encoded_path_segment(segment):
+    return is_valid_encoded_path_segment.regex.match(segment) is not None
+
+
+@static_vars(regex=re.compile(
+    r'^([\w%s]|(%s))*$' % (re.escape('-.~:@!$&\'()*+,;/?'), PERCENT_REGEX)))
+def is_valid_encoded_query_key(key):
+    return is_valid_encoded_query_key.regex.match(key) is not None
+
+
+@static_vars(regex=re.compile(
+    r'^([\w%s]|(%s))*$' % (re.escape('-.~:@!$&\'()*+,;/?='), PERCENT_REGEX)))
+def is_valid_encoded_query_value(value):
+    return is_valid_encoded_query_value.regex.match(value) is not None
+
+
+@static_vars(regex=re.compile(r'[a-zA-Z][a-zA-Z\-\.\+]*'))
+def is_valid_scheme(scheme):
+    return is_valid_scheme.regex.match(scheme) is not None
+
+
+@static_vars(regex=re.compile('[%s]' % re.escape(INVALID_DOMAIN_CHARS)))
+def is_valid_domain(domain):
+    toks = domain.split('.')
+    if toks[-1] == '':  # Trailing '.' in a fully qualified domain name.
+        toks.pop()
+
+    for tok in toks:
+        if is_valid_domain.regex.search(tok) is not None:
+            return False
+
+    return '' not in toks  # Adjacent periods aren't allowed.
+
+
+def _get_scheme(url):
+    if url.lstrip().startswith('//'):  # Protocol relative URL.
+        return ''
+
+    before_colon = url[:max(0, url.find(':'))]
+    if before_colon in COLON_SEPARATED_SCHEMES:
+        scheme = before_colon
+    else:
+        scheme = url[:max(0, url.find('://'))] or None
+    return scheme if (scheme is not None and is_valid_scheme(scheme)) else None
+
+
+def _set_scheme(url, newscheme):
+    scheme = _get_scheme(url)
+    newscheme = newscheme or ''
+    newseparator = ':' if newscheme in COLON_SEPARATED_SCHEMES else '://'
+    if scheme == '':  # Protocol relative URL.
+        url = '%s:%s' % (newscheme, url)
+    elif scheme is None and url:  # No scheme.
+        url = ''.join([newscheme, newseparator, url])
+    elif scheme:  # Existing scheme.
+        remainder = url[len(scheme):]
+        if remainder.startswith('://'):
+            remainder = remainder[3:]
+        elif remainder.startswith(':'):
+            remainder = remainder[1:]
+        url = ''.join([newscheme, newseparator, remainder])
+    return url
+
+
+def urlsplit(url):
+    """
+    Parameters:
+      url: URL string to split.
+
+    Returns: urlparse.SplitResult tuple subclass, just like
+    urlparse.urlsplit() returns, with fields (scheme, netloc, path,
+    query, fragment, username, password, hostname, port). See the url
+    below for more details on urlsplit().
+
+      http://docs.python.org/library/urlparse.html#urlparse.urlsplit
+    """
+    original_scheme = _get_scheme(url)
+
+    def _change_urltoks_scheme(tup, scheme):
+        l = list(tup)
+        l[0] = scheme
+        return tuple(l)
+
+    # urlsplit() only parses the query for schemes in urlparse.uses_query,
+    # so switch to 'http' (a scheme in urlparse.uses_query) for
+    # urlparse.urlsplit() and switch back afterwards.
+    if original_scheme is not None:
+        url = _set_scheme(url, 'http')
+    toks = urllib.parse.urlsplit(url)
+    toks_orig_scheme = _change_urltoks_scheme(toks, original_scheme)
+    return urllib.parse.SplitResult(*toks_orig_scheme)
+
+
+def urljoin(base, url):
+    """
+    Parameters:
+      base: Base URL to join with <url>.
+      url: Relative or absolute URL to join with <base>.
+
+    Returns: The resultant URL from joining <base> and <url>.
+    """
+    base_scheme, url_scheme = urlsplit(base).scheme, urlsplit(url).scheme
+    httpbase = _set_scheme(base, 'http')
+    joined = urllib.parse.urljoin(httpbase, url)
+    if not url_scheme:
+        joined = _set_scheme(joined, base_scheme)
+    return joined
+
+
+def join_path_segments(*args):
+    """
+    Join multiple lists of path segments together, intelligently
+    handling path segments borders to preserve intended slashes of the
+    final constructed path.
+
+    This function is not encoding aware. It doesn't test for, or change,
+    the encoding of path segments it is passed.
+
+    Examples:
+      join_path_segments(['a'], ['b']) == ['a','b']
+      join_path_segments(['a',''], ['b']) == ['a','b']
+      join_path_segments(['a'], ['','b']) == ['a','b']
+      join_path_segments(['a',''], ['','b']) == ['a','','b']
+      join_path_segments(['a','b'], ['c','d']) == ['a','b','c','d']
+
+    Returns: A list containing the joined path segments.
+    """
+    finals = []
+    for segments in args:
+        if not segments or segments == ['']:
+            continue
+        elif not finals:
+            finals.extend(segments)
+        else:
+            # Example #1: ['a',''] + ['b'] == ['a','b']
+            # Example #2: ['a',''] + ['','b'] == ['a','','b']
+            if finals[-1] == '' and (segments[0] != '' or len(segments) > 1):
+                finals.pop(-1)
+            # Example: ['a'] + ['','b'] == ['a','b']
+            elif finals[-1] != '' and segments[0] == '' and len(segments) > 1:
+                segments = segments[1:]
+            finals.extend(segments)
+    return finals
+
+
+def remove_path_segments(segments, remove):
+    """
+    Removes the path segments of <remove> from the end of the path
+    segments <segments>.
+
+    Examples:
+      # '/a/b/c' - 'b/c' == '/a/'
+      remove_path_segments(['','a','b','c'], ['b','c']) == ['','a','']
+      # '/a/b/c' - '/b/c' == '/a'
+      remove_path_segments(['','a','b','c'], ['','b','c']) == ['','a']
+
+    Returns: The list of all remaining path segments after the segments
+    in <remove> have been removed from the end of <segments>. If no
+    segments from <remove> were removed from <segments>, <segments> is
+    returned unmodified.
+    """
+    # [''] means a '/', which is properly represented by ['', ''].
+    if segments == ['']:
+        segments.append('')
+    if remove == ['']:
+        remove.append('')
+
+    ret = None
+    if remove == segments:
+        ret = []
+    elif len(remove) > len(segments):
+        ret = segments
+    else:
+        toremove = list(remove)
+
+        if len(remove) > 1 and remove[0] == '':
+            toremove.pop(0)
+
+        if toremove and toremove == segments[-1 * len(toremove):]:
+            ret = segments[:len(segments) - len(toremove)]
+            if remove[0] != '' and ret:
+                ret.append('')
+        else:
+            ret = segments
+
+    return ret
 
 
 class Path(object):
@@ -261,7 +545,7 @@ class Path(object):
                          (path, self._path_from_segments(segments)))
                     warnings.warn(s, UserWarning)
             segments.append(utf8(segment))
-        
+
         # In Python 3, utf8() returns Bytes objects that must be decoded
         # into strings before they can be passed to urllib.unquote(). In
         # Python 2, utf8() returns strings that can be passed directly
@@ -280,10 +564,9 @@ class Path(object):
 
         Returns: A path string with quoted path segments.
         """
-        if '%' not in ''.join(segments):  # Don't double-encode the path.
-            segments = [
-                quote(utf8(attemptstr(segment)), self.SAFE_SEGMENT_CHARS)
-                for segment in segments]
+        segments = [
+            quote(utf8(attemptstr(segment)), self.SAFE_SEGMENT_CHARS)
+            for segment in segments]
         return '/'.join(segments)
 
 
@@ -518,28 +801,40 @@ class Query(object):
         for key, value in items:
             self._params.add(key, value)
 
-    def encode(self, delimeter='&'):
+    def encode(self, delimiter='&', quote_plus=True, delimeter=_absent):
         """
         Examples:
           Query('a=a&b=#').encode() == 'a=a&b=%23'
           Query('a=a&b=#').encode(';') == 'a=a;b=%23'
+          Query('a+b=c+d').encode(quote_plus=False) == 'a%20b=c%20d'
 
-        Returns: A URL encoded query string using <delimeter> as the
-        delimeter separating key:value pairs. The most common and
-        default delimeter is '&', but ';' can also be specified. ';' is
-        W3C recommended.
+        Until furl v0.4.6, the 'delimiter' argument was incorrectly
+        spelled 'delimeter'. For backwards compatibility, accept both
+        the correct 'delimiter' and the old, mispelled 'delimeter'.
+
+        Returns: A URL encoded query string using <delimiter> as the
+        delimiter separating key:value pairs. The most common and
+        default delimiter is '&', but ';' can also be specified. ';' is
+        W3C recommended. Parameter keys and values are encoded
+        application/x-www-form-urlencoded if <quote_plus> is True,
+        percent-encoded otherwise.
         """
+        if delimeter is not _absent:
+            delimiter = delimeter
+
         pairs = []
+        sixurl = urllib.parse  # six.moves.urllib.parse
+        quote_func = sixurl.quote_plus if quote_plus else sixurl.quote
         for key, value in self.params.iterallitems():
             utf8key = utf8(key, utf8(attemptstr(key)))
             utf8value = utf8(value, utf8(attemptstr(value)))
-            quoted_key = quote_plus(utf8key, self.SAFE_KEY_CHARS)
-            quoted_value = quote_plus(utf8value, self.SAFE_VALUE_CHARS)
+            quoted_key = quote_func(utf8key, self.SAFE_KEY_CHARS)
+            quoted_value = quote_func(utf8value, self.SAFE_VALUE_CHARS)
             pair = '='.join([quoted_key, quoted_value])
             if value is None:  # Example: http://sprop.su/?param
                 pair = quoted_key
             pairs.append(pair)
-        return delimeter.join(pairs)
+        return delimiter.join(pairs)
 
     def __eq__(self, other):
         return str(self) == str(other)
@@ -865,13 +1160,13 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
         """
         Parse and load a URL.
 
-        Raises: ValueError on invalid URL (for example malformed IPv6
-        address or invalid port).
+        Raises: ValueError on invalid URL, like a malformed IPv6 address
+        or invalid port.
         """
+        self.username = self.password = None
         self._host = self._port = self._scheme = None
-        self.username = self.password = self.scheme = None
 
-        if not isinstance(url, str):
+        if not isinstance(url, six.string_types):
             url = str(url)
 
         # urlsplit() raises a ValueError on malformed IPv6 addresses in
@@ -905,21 +1200,39 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
     @host.setter
     def host(self, host):
         """
-        Raises: ValueError on malformed IPv6 address.
+        Raises: ValueError on invalid host or malformed IPv6 address.
         """
+        # Invalid IPv6 literal.
         urllib.parse.urlsplit('http://%s/' % host)  # Raises ValueError.
+
+        # Invalid host string.
+        resembles_ipv6_literal = (
+            host is not None and lget(host, 0) == '[' and ':' in host and
+            lget(host, -1) == ']')
+        if (host is not None and not resembles_ipv6_literal and
+           not is_valid_domain(host)):
+            errmsg = (
+                "Invalid host '%s'. Host strings must have at least one "
+                "non-period character, can't contain any of '%s', and can't "
+                "have adjacent periods.")
+            raise ValueError(errmsg % (host, INVALID_DOMAIN_CHARS))
+
+        if callable_attr(host, 'lower'):
+            host = host.lower()
+        if callable_attr(host, 'startswith') and host.startswith('xn--'):
+            host = idna_decode(host)
         self._host = host
 
     @property
     def port(self):
-        return self._port
+        return self._port or DEFAULT_PORTS.get(self.scheme)
 
     @port.setter
     def port(self, port):
         """
-        A port value can 1-65535 or None meaning no port specified. If <port>
-        is None and self.scheme is a known scheme in DEFAULT_PORTS, the default
-        port value from DEFAULT_PORTS will be used.
+        The port value can be 1-65535 or None, meaning no port specified. If
+        <port> is None and self.scheme is a known scheme in DEFAULT_PORTS,
+        the default port value from DEFAULT_PORTS will be used.
 
         Raises: ValueError on invalid port.
         """
@@ -928,21 +1241,21 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
         elif is_valid_port(port):
             self._port = int(str(port))
         else:
-            raise ValueError("Invalid port: '%s'" % port)
+            raise ValueError("Invalid port '%s'." % port)
 
     @property
     def netloc(self):
-        userpass = self.username or ''
+        userpass = quote(self.username or '', safe='')
         if self.password is not None:
-            userpass += ':' + self.password
+            userpass += ':' + quote(self.password, safe='')
         if userpass or self.username is not None:
             userpass += '@'
 
-        netloc = self.host or ''
+        netloc = idna_encode(self.host) or ''
         if self.port and self.port != DEFAULT_PORTS.get(self.scheme):
             netloc += ':' + str(self.port)
 
-        netloc = ((userpass or '') + (netloc or ''))
+        netloc = (userpass or '') + (netloc or '')
         return netloc if (netloc or self.host == '') else None
 
     @netloc.setter
@@ -950,7 +1263,7 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
         """
         Params:
           netloc: Network location string, like 'google.com' or
-            'google.com:99'.
+            'user:pass@google.com:99'.
         Raises: ValueError on invalid port or malformed IPv6 address.
         """
         # Raises ValueError on malformed IPv6 addresses.
@@ -970,28 +1283,51 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
             if ']' in netloc:
                 colonpos, bracketpos = netloc.rfind(':'), netloc.rfind(']')
                 if colonpos > bracketpos and colonpos != bracketpos + 1:
-                    raise ValueError("Invalid netloc: '%s'" % netloc)
+                    raise ValueError("Invalid netloc '%s'." % netloc)
                 elif colonpos > bracketpos and colonpos == bracketpos + 1:
                     host, port = netloc.rsplit(':', 1)
                 else:
-                    host = netloc.lower()
+                    host = netloc
             else:
                 host, port = netloc.rsplit(':', 1)
-                host = host.lower()
+                host = host
         else:
-            host = netloc.lower()
+            host = netloc
 
         # Avoid side effects by assigning self.port before self.host so
         # that if an exception is raised when assigning self.port,
         # self.host isn't updated.
         self.port = port  # Raises ValueError on invalid port.
         self.host = host or None
-        self.username = username or None
-        self.password = password or None
+        self.username = None if username is None else unquote(username)
+        self.password = None if password is None else unquote(password)
+
+    @property
+    def origin(self):
+        port = ''
+        scheme = self.scheme or ''
+        host = idna_encode(self.host) or ''
+        if self.port and self.port != DEFAULT_PORTS.get(self.scheme):
+            port = ':%s' % self.port
+        origin = '%s://%s%s' % (scheme, host, port)
+        return origin
+
+    @origin.setter
+    def origin(self, origin):
+        toks = origin.split('://', 1)
+        if len(toks) == 1:
+            host_port = origin
+        else:
+            self.scheme, host_port = toks
+
+        if ':' in host_port:
+            self.host, self.port = host_port.split(':', 1)
+        else:
+            self.host = host_port
 
     @property
     def url(self):
-        return str(self)
+        return self.tostr()
 
     @url.setter
     def url(self, url):
@@ -1046,18 +1382,18 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
         return self
 
     def set(self, args=_absent, path=_absent, fragment=_absent, scheme=_absent,
-            netloc=_absent, fragment_path=_absent, fragment_args=_absent,
-            fragment_separator=_absent, host=_absent, port=_absent,
-            query=_absent, query_params=_absent, username=_absent,
-            password=_absent):
-        """
-        Set components of a url and return this furl instance, <self>.
+            netloc=_absent, origin=_absent, fragment_path=_absent,
+            fragment_args=_absent, fragment_separator=_absent, host=_absent,
+            port=_absent, query=_absent, query_params=_absent,
+            username=_absent, password=_absent):
+        """Set components of a url and return this furl instance, <self>.
 
         If any overlapping, and hence possibly conflicting, parameters
         are provided, appropriate UserWarning's will be raised. The
         groups of parameters that could potentially overlap are
 
-          <netloc> and (<host> or <port>)
+          <scheme> and <origin>
+          <origin>, <netloc>, and/or (<host> or <port>)
           <fragment> and (<fragment_path> and/or <fragment_args>)
           any two or all of <query>, <args>, and/or <query_params>
 
@@ -1081,6 +1417,7 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
           fragment: Fragment string to adopt.
           scheme: Scheme string to adopt.
           netloc: Network location string to adopt.
+          origin: Scheme and netloc.
           query: Query string to adopt.
           query_params: A dictionary of query keys and values or list of
             key:value items to adopt.
@@ -1097,73 +1434,81 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
           password: Password string to adopt.
         Raises:
           ValueError on invalid port.
-          UserWarning if <netloc> and (<host> and/or <port>) are
-            provided.
-          UserWarning if <query>, <args>, and/or <query_params> are
+          UserWarning if <scheme> and <origin> are provided.
+          UserWarning if <origin>, <netloc> and/or (<host> and/or <port>) are
             provided.
+          UserWarning if <query>, <args>, and/or <query_params> are provided.
           UserWarning if <fragment> and (<fragment_path>,
             <fragment_args>, and/or <fragment_separator>) are provided.
         Returns: <self>.
         """
-        netloc_present = netloc is not _absent
-        if (netloc_present and (host is not _absent or port is not _absent)):
-            s = ('Possible parameter overlap: <netloc> and <host> and/or '
-                 '<port> provided. See furl.set() documentation for more '
-                 'details.')
+        def present(v):
+            return v is not _absent
+
+        if present(scheme) and present(origin):
+            s = ('Possible parameter overlap: <scheme> and <origin>. See '
+                 'furl.set() documentation for more details.')
             warnings.warn(s, UserWarning)
-        if ((args is not _absent and query is not _absent) or
-            (query is not _absent and query_params is not _absent) or
-                (args is not _absent and query_params is not _absent)):
+        l = [present(netloc), present(origin), present(host) or present(port)]
+        if sum(l) >= 2:
+            s = ('Possible parameter overlap: <origin>, <netloc> and/or '
+                 '(<host> and/or <port>) provided. See furl.set() '
+                 'documentation for more details.')
+            warnings.warn(s, UserWarning)
+        if sum(present(p) for p in [args, query, query_params]) >= 2:
             s = ('Possible parameter overlap: <query>, <args>, and/or '
                  '<query_params> provided. See furl.set() documentation for '
                  'more details.')
             warnings.warn(s, UserWarning)
-        if (fragment is not _absent and
-            (fragment_path is not _absent or fragment_args is not _absent or
-             (fragment_separator is not _absent))):
+        l = [fragment_path, fragment_args, fragment_separator]
+        if present(fragment) and any(present(p) for p in l):
             s = ('Possible parameter overlap: <fragment> and '
                  '(<fragment_path>and/or <fragment_args>) or <fragment> '
                  'and <fragment_separator> provided. See furl.set() '
                  'documentation for more details.')
             warnings.warn(s, UserWarning)
 
-        # Avoid side effects if exceptions are raised.
-        oldnetloc, oldport = self.netloc, self.port
+        # Guard against side effects on exception.
+        original_url = self.url
         try:
+            if username is not _absent:
+                self.username = username
+            if password is not _absent:
+                self.password = password
             if netloc is not _absent:
                 # Raises ValueError on invalid port or malformed IP.
                 self.netloc = netloc
+            if origin is not _absent:
+                # Raises ValueError on invalid port or malformed IP.
+                self.origin = origin
+            if scheme is not _absent:
+                self.scheme = scheme
+            if host is not _absent:
+                # Raises ValueError on invalid host or malformed IP.
+                self.host = host
             if port is not _absent:
                 self.port = port  # Raises ValueError on invalid port.
-        except ValueError:
-            self.netloc, self.port = oldnetloc, oldport
+
+            if path is not _absent:
+                self.path.load(path)
+            if query is not _absent:
+                self.query.load(query)
+            if args is not _absent:
+                self.query.load(args)
+            if query_params is not _absent:
+                self.query.load(query_params)
+            if fragment is not _absent:
+                self.fragment.load(fragment)
+            if fragment_path is not _absent:
+                self.fragment.path.load(fragment_path)
+            if fragment_args is not _absent:
+                self.fragment.query.load(fragment_args)
+            if fragment_separator is not _absent:
+                self.fragment.separator = fragment_separator
+        except:
+            self.load(original_url)
             raise
 
-        if username is not _absent:
-            self.username = username
-        if password is not _absent:
-            self.password = password
-        if scheme is not _absent:
-            self.scheme = scheme
-        if host is not _absent:
-            self.host = host
-
-        if path is not _absent:
-            self.path.load(path)
-        if query is not _absent:
-            self.query.load(query)
-        if args is not _absent:
-            self.query.load(args)
-        if query_params is not _absent:
-            self.query.load(query_params)
-        if fragment is not _absent:
-            self.fragment.load(fragment)
-        if fragment_path is not _absent:
-            self.fragment.path.load(fragment_path)
-        if fragment_args is not _absent:
-            self.fragment.query.load(fragment_args)
-        if fragment_separator is not _absent:
-            self.fragment.separator = fragment_separator
         return self
 
     def remove(self, args=_absent, path=_absent, fragment=_absent,
@@ -1218,34 +1563,12 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
             self.fragment.query.remove(fragment_args)
         return self
 
-    def join(self, url):
-        self.load(urljoin(self.url, str(url)))
-        return self
-
-    def copy(self):
-        return self.__class__(self)
-
-    def __eq__(self, other):
-        try:
-            return self.url == other.url
-        except AttributeError:
-            return None
-
-    def __ne__(self, other):
-        return not self == other
-
-    def __setattr__(self, attr, value):
-        if (not PathCompositionInterface.__setattr__(self, attr, value) and
-            not QueryCompositionInterface.__setattr__(self, attr, value) and
-            not FragmentCompositionInterface.__setattr__(self, attr, value)):
-            object.__setattr__(self, attr, value)
-
-    def __unicode__(self):
+    def tostr(self, query_delimiter='&', query_quote_plus=True):
         url = urllib.parse.urlunsplit((
             self.scheme or '',  # Must be text type in Python 3.
             self.netloc,
             str(self.path),
-            str(self.query),
+            self.query.encode(query_delimiter, query_quote_plus),
             str(self.fragment),
         ))
 
@@ -1262,238 +1585,35 @@ class furl(URLPathCompositionInterface, QueryCompositionInterface,
               (url == '' or  # Protocol relative URL.
                (url == '%s:' % self.scheme and not str(self.path)))):
             url += '//'
-        return url
+        return str(url)
+
+    def join(self, *urls):
+        for url in urls:
+            if not isinstance(url, six.string_types):
+                url = str(url)
+            self.load(urljoin(self.url, url))
+        return self
+
+    def copy(self):
+        return self.__class__(self)
+
+    def __eq__(self, other):
+        try:
+            return self.url == other.url
+        except AttributeError:
+            return None
+
+    def __ne__(self, other):
+        return not self == other
+
+    def __setattr__(self, attr, value):
+        if (not PathCompositionInterface.__setattr__(self, attr, value) and
+           not QueryCompositionInterface.__setattr__(self, attr, value) and
+           not FragmentCompositionInterface.__setattr__(self, attr, value)):
+            object.__setattr__(self, attr, value)
+
+    def __unicode__(self):
+        return self.tostr()
 
     def __repr__(self):
         return "%s('%s')" % (self.__class__.__name__, str(self))
-
-
-def _get_scheme(url):
-    if url.lstrip().startswith('//'):  # Protocol relative URL.
-        return ''
-    beforeColon = url[:max(0, url.find(':'))]
-    if beforeColon in COLON_SEPARATED_SCHEMES:
-        return beforeColon
-    return url[:max(0, url.find('://'))] or None
-
-
-def _set_scheme(url, newscheme):
-    scheme = _get_scheme(url)
-    newscheme = newscheme or ''
-    newseparator = ':' if newscheme in COLON_SEPARATED_SCHEMES else '://'
-    if scheme == '':  # Protocol relative URL.
-        url = '%s:%s' % (newscheme, url)
-    elif scheme is None and url:  # No scheme.
-        url = ''.join([newscheme, newseparator, url])
-    elif scheme:  # Existing scheme.
-        remainder = url[len(scheme):]
-        if remainder.startswith('://'):
-            remainder = remainder[3:]
-        elif remainder.startswith(':'):
-            remainder = remainder[1:]
-        url = ''.join([newscheme, newseparator, remainder])
-    return url
-
-
-def urlsplit(url):
-    """
-    Parameters:
-      url: URL string to split.
-
-    Returns: urlparse.SplitResult tuple subclass, just like
-    urlparse.urlsplit() returns, with fields (scheme, netloc, path,
-    query, fragment, username, password, hostname, port). See the url
-    below for more details on urlsplit().
-
-      http://docs.python.org/library/urlparse.html#urlparse.urlsplit
-    """
-    original_scheme = _get_scheme(url)
-
-    def _change_urltoks_scheme(tup, scheme):
-        l = list(tup)
-        l[0] = scheme
-        return tuple(l)
-
-    # urlsplit() only parses the query for schemes in urlparse.uses_query,
-    # so switch to 'http' (a scheme in urlparse.uses_query) for
-    # urlparse.urlsplit() and switch back afterwards.
-    if original_scheme is not None:
-        url = _set_scheme(url, 'http')
-    toks = urllib.parse.urlsplit(url)
-    toks_orig_scheme = _change_urltoks_scheme(toks, original_scheme)
-    return urllib.parse.SplitResult(*toks_orig_scheme)
-
-
-def urljoin(base, url):
-    """
-    Parameters:
-      base: Base URL to join with <url>.
-      url: Relative or absolute URL to join with <base>.
-
-    Returns: The resultant URL from joining <base> and <url>.
-    """
-    base_scheme, url_scheme = urlsplit(base).scheme, urlsplit(url).scheme
-    httpbase = _set_scheme(base, 'http')
-    joined = urllib.parse.urljoin(httpbase, url)
-    if not url_scheme:
-        joined = _set_scheme(joined, base_scheme)
-    return joined
-
-
-def join_path_segments(*args):
-    """
-    Join multiple lists of path segments together, intelligently
-    handling path segments borders to preserve intended slashes of the
-    final constructed path.
-
-    This function is not encoding aware - it does not test for or change
-    the encoding of path segments it is passed.
-
-    Examples:
-      join_path_segments(['a'], ['b']) == ['a','b']
-      join_path_segments(['a',''], ['b']) == ['a','b']
-      join_path_segments(['a'], ['','b']) == ['a','b']
-      join_path_segments(['a',''], ['','b']) == ['a','','b']
-      join_path_segments(['a','b'], ['c','d']) == ['a','b','c','d']
-
-    Returns: A list containing the joined path segments.
-    """
-    finals = []
-    for segments in args:
-        if not segments or segments == ['']:
-            continue
-        elif not finals:
-            finals.extend(segments)
-        else:
-            # Example #1: ['a',''] + ['b'] == ['a','b']
-            # Example #2: ['a',''] + ['','b'] == ['a','','b']
-            if finals[-1] == '' and (segments[0] != '' or len(segments) > 1):
-                finals.pop(-1)
-            # Example: ['a'] + ['','b'] == ['a','b']
-            elif finals[-1] != '' and segments[0] == '' and len(segments) > 1:
-                segments = segments[1:]
-            finals.extend(segments)
-    return finals
-
-
-def remove_path_segments(segments, remove):
-    """
-    Removes the path segments of <remove> from the end of the path
-    segments <segments>.
-
-    Examples:
-      # '/a/b/c' - 'b/c' == '/a/'
-      remove_path_segments(['','a','b','c'], ['b','c']) == ['','a','']
-      # '/a/b/c' - '/b/c' == '/a'
-      remove_path_segments(['','a','b','c'], ['','b','c']) == ['','a']
-
-    Returns: The list of all remaining path segments after the segments
-    in <remove> have been removed from the end of <segments>. If no
-    segments from <remove> were removed from <segments>, <segments> is
-    returned unmodified.
-    """
-    # [''] means a '/', which is properly represented by ['', ''].
-    if segments == ['']:
-        segments.append('')
-    if remove == ['']:
-        remove.append('')
-
-    ret = None
-    if remove == segments:
-        ret = []
-    elif len(remove) > len(segments):
-        ret = segments
-    else:
-        toremove = list(remove)
-
-        if len(remove) > 1 and remove[0] == '':
-            toremove.pop(0)
-
-        if toremove and toremove == segments[-1 * len(toremove):]:
-            ret = segments[:len(segments) - len(toremove)]
-            if remove[0] != '' and ret:
-                ret.append('')
-        else:
-            ret = segments
-
-    return ret
-
-
-def attemptstr(o):
-    try:
-        return str(o)
-    except:
-        return o
-
-
-def utf8(o, default=_absent):
-    try:
-        return o.encode('utf8')
-    except:
-        return o if default is _absent else default
-
-
-def is_valid_port(port):
-    port = str(port)
-    if not port.isdigit() or int(port) == 0 or int(port) > 65535:
-        return False
-    return True
-
-
-def callable_attr(obj, attr):
-    return hasattr(obj, attr) and callable(getattr(obj, attr))
-
-
-#
-# TODO(grun): These regex functions need to be expanded to reflect the
-# fact that the valid encoding for a URL Path segment is different from
-# a Fragment Path segment, and valid URL Query key and value encoding
-# is different than valid Fragment Query key and value encoding.
-#
-# For example, '?' and '#' don't need to be encoded in Fragment Path
-# segments but they must be encoded in URL Path segments.
-#
-# Similarly, '#' doesn't need to be encoded in Fragment Query keys and
-# values, but must be encoded in URL Query keys and values.
-#
-# Perhaps merge them with URLPath, FragmentPath, URLQuery, and
-# FragmentQuery when those new classes are created (see the TODO
-# currently at the top of the source, 02/03/2012).
-#
-
-# RFC 3986
-#   unreserved  = ALPHA / DIGIT / "-" / "." / "_" / "~"
-#
-#   pct-encoded = "%" HEXDIG HEXDIG
-#
-#   sub-delims  = "!" / "$" / "&" / "'" / "(" / ")"
-#                 / "*" / "+" / "," / ";" / "="
-#
-#   pchar         = unreserved / pct-encoded / sub-delims / ":" / "@"
-#
-#   ====
-#   Path
-#   ====
-#   segment       = *pchar
-#
-#   =====
-#   Query
-#   =====
-#   query       = *( pchar / "/" / "?" )
-#
-VALID_ENCODED_PATH_SEGMENT_REGEX = re.compile(
-    r'^([\w\-\.\~\:\@\!\$\&\'\(\)\*\+\,\;\=]|(\%[\da-fA-F][\da-fA-F]))*$')
-def is_valid_encoded_path_segment(segment):
-    return bool(VALID_ENCODED_PATH_SEGMENT_REGEX.match(segment))
-
-
-VALID_ENCODED_QUERY_KEY_REGEX = re.compile(
-    r'^([\w\-\.\~\:\@\!\$\&\'\(\)\*\+\,\;\/\?]|(\%[\da-fA-F][\da-fA-F]))*$')
-def is_valid_encoded_query_key(key):
-    return bool(VALID_ENCODED_QUERY_KEY_REGEX.match(key))
-
-
-VALID_ENCODED_QUERY_VALUE_REGEX = re.compile(
-    r'^([\w\-\.\~\:\@\!\$\&\'\(\)\*\+\,\;\/\?\=]|(\%[\da-fA-F][\da-fA-F]))*$')
-def is_valid_encoded_query_value(value):
-    return bool(VALID_ENCODED_QUERY_VALUE_REGEX.match(value))
