@@ -1,4 +1,5 @@
 # -*- coding: utf-8 -*-
+
 #
 # furl - URL manipulation made simple.
 #
@@ -13,15 +14,19 @@ import warnings
 from abc import ABCMeta, abstractmethod
 
 import six
-from six.moves import urllib
+from six.moves import zip, urllib
 
 import furl
-from furl.compat import unittest
 from furl.omdict1D import omdict1D
 from furl.compat import OrderedDict as odict
 
 PYTHON_27PLUS = sys.version_info >= (2, 7)
 
+if PYTHON_27PLUS:
+    import unittest
+else:
+    import unittest2 as unittest
+
 
 #
 # TODO(grun): Add tests for furl objects with strict=True. Make sure
@@ -98,7 +103,7 @@ class itemstr(str, itemcontainer):
         parsed = urllib.parse.parse_qsl(self, keep_blank_values=True)
         pairstrs = [s2 for s1 in self.split('&')
                     for s2 in s1.split(';')]
-        for (key, value), pairstr in six.moves.zip(parsed, pairstrs):
+        for (key, value), pairstr in zip(parsed, pairstrs):
             if key == urllib.parse.quote_plus(pairstr):
                 value = None
             items.append((key, value))
@@ -178,13 +183,13 @@ class TestPath(unittest.TestCase):
         assert str(p) == '/a/b/c/'
 
     def test_encoding(self):
-        encoded = ['a%20a', '/%7haypepps/', 'a/:@/a', 'a%2Fb']
-        unencoded = ['a+a', '/~haypepps/', 'a/:@/a', 'a/b']
+        decoded = ['a+a', '/#haypepps/', 'a/:@/a', 'a/b']
+        encoded = ['a%20a', '/%23haypepps/', 'a/:@/a', 'a%2Fb']
 
         for path in encoded:
             assert str(furl.Path(path)) == path
 
-        for path in unencoded:
+        for path in decoded:
             assert str(furl.Path(path)) == urllib.parse.quote(
                 path, "/:@-._~!$&'()*+,;=")
 
@@ -488,8 +493,8 @@ class TestPath(unittest.TestCase):
         assert p
 
     def test_unicode(self):
-        paths = ['/wiki/Восход', u'/wiki/Восход']
-        path_encoded = '/wiki/%D0%92%D0%BE%D1%81%D1%85%D0%BE%D0%B4'
+        paths = ['/wiki/ロリポップ', u'/wiki/ロリポップ']
+        path_encoded = '/wiki/%E3%83%AD%E3%83%AA%E3%83%9D%E3%83%83%E3%83%97'
         for path in paths:
             p = furl.Path(path)
             assert str(p) == path_encoded
@@ -544,7 +549,7 @@ class TestQuery(unittest.TestCase):
             'a=a&a=a', 'space=a+a&space=b+b',
             # Empty keys and/or values.
             '=', 'a=', 'a=a&a=', '=a&=b',
-            # Semicolon delimeter, like 'a=a;b=b'.
+            # Semicolon delimiter, like 'a=a;b=b'.
             'a=a;a=a', 'space=a+a;space=b+b',
         ]))
         self.items = (self.itemlists + self.itemdicts + self.itemomdicts +
@@ -556,8 +561,10 @@ class TestQuery(unittest.TestCase):
             assert q.params.allitems() == items.allitems()
             pairs = ['%s=%s' % (i[0], i[1]) for i in self._quote_items(items)]
 
-            # encode() and __str__().
-            assert str(q) == q.encode() == q.encode('&')
+            # encode() accepts both 'delimiter' and 'delimeter'. The
+            # latter was incorrectly used until furl v0.4.6.
+            e = q.encode
+            assert e(';') == e(delimiter=';') == e(delimeter=';')
 
             # __nonzero__().
             if items.allitems():
@@ -671,7 +678,7 @@ class TestQuery(unittest.TestCase):
         q = furl.Query('a=&=b')
         assert q.params == {'a': '', '': 'b'} and str(q) == 'a=&=b'
 
-        # ';' is a valid query delimeter.
+        # ';' is a valid query delimiter.
         q = furl.Query('=;=')
         assert q.params.allitems() == [('', ''), ('', '')] and str(q) == '=&='
         q = furl.Query('a=a;b=b;c=')
@@ -706,7 +713,7 @@ class TestQuery(unittest.TestCase):
             q.params = items.original()
             assert isinstance(q.params, omdict1D)
 
-            pairs = six.moves.zip(q.params.iterallitems(), items.iterallitems())
+            pairs = zip(q.params.iterallitems(), items.iterallitems())
             for item1, item2 in pairs:
                 assert item1 == item2
 
@@ -721,15 +728,15 @@ class TestQuery(unittest.TestCase):
         assert str(q) == 'prp=&slrp=' and q.params['slrp'] == ''
 
     def test_unicode(self):
-        pairs = [('Восход', 'testä'), (u'Восход', u'testä')]
-        key_encoded = '%D0%92%D0%BE%D1%81%D1%85%D0%BE%D0%B4'
+        pairs = [('ロリポップ', 'testä'), (u'ロリポップ', u'testä')]
+        key_encoded = '%E3%83%AD%E3%83%AA%E3%83%9D%E3%83%83%E3%83%97'
         value_encoded = 'test%C3%A4'
 
         for key, value in pairs:
             q = furl.Query('%s=%s' % (key, value))
             assert q.params[key] == value
             assert str(q) == '%s=%s' % (key_encoded, value_encoded)
-            
+
             q = furl.Query()
             q.params[key] = value
             assert q.params[key] == value
@@ -745,6 +752,18 @@ class TestQuery(unittest.TestCase):
         assert q1 == q11 and str(q1) == str(q11)
         assert q1 != q2 and str(q1) != str(q2)
 
+    def test_encode(self):
+        for items in self.items:
+            q = furl.Query(items.original())
+            # encode() and __str__().
+            assert str(q) == q.encode() == q.encode('&')
+
+        # Accept both percent-encoded ('a=b%20c') and
+        # application/x-www-form-urlencoded ('a=b+c') pairs as input.
+        query = furl.Query('a=b%20c&d=e+f')
+        assert query.encode(';') == 'a=b+c;d=e+f'
+        assert query.encode(';', quote_plus=False) == 'a=b%20c;d=e%20f'
+
     def _quote_items(self, items):
         # Calculate the expected querystring with proper query encoding.
         #   Valid query key characters: "/?:@-._~!$'()*,;"
@@ -836,8 +855,8 @@ class TestFragment(unittest.TestCase):
 
     def test_add(self):
         f = furl.Fragment('')
-        assert f is f.add(path='one two three', args=[('a', 'a'), ('s', 's s')])
-        assert str(f) == 'one%20two%20three?a=a&s=s+s'
+        assert f is f.add(path='one two', args=[('a', 'a'), ('s', 's s')])
+        assert str(f) == 'one%20two?a=a&s=s+s'
 
         f = furl.Fragment('break?legs=broken')
         assert f is f.add(path='horse bones', args=[('a', 'a'), ('s', 's s')])
@@ -845,8 +864,8 @@ class TestFragment(unittest.TestCase):
 
     def test_set(self):
         f = furl.Fragment('asdf?lol=sup&foo=blorp')
-        assert f is f.set(path='one two three', args=[('a', 'a'), ('s', 's s')])
-        assert str(f) == 'one%20two%20three?a=a&s=s+s'
+        assert f is f.set(path='one two', args=[('a', 'a'), ('s', 's s')])
+        assert str(f) == 'one%20two?a=a&s=s+s'
 
         assert f is f.set(path='!', separator=False)
         assert f.separator is False
@@ -886,13 +905,14 @@ class TestFragment(unittest.TestCase):
         f = furl.furl()
         f.fragment = 'a?b?c?d?'
         assert f.url == '#a?b?c?d?'
-        # TODO(grun): Once encoding has been fixed with URLPath and
-        # FragmentPath, the below line should be:
-        #
-        #  assert str(f.fragment) == str(f.path) == 'a?b?c?d?'
-        #
         assert str(f.fragment) == 'a?b?c?d?'
 
+    def test_unicode(self):
+        for fragment in ['ロリポップ', u'ロリポップ']:
+            f = furl.furl('http://sprop.ru/#ja').set(fragment=fragment)
+            assert str(f.fragment) == (
+                '%E3%83%AD%E3%83%AA%E3%83%9D%E3%83%83%E3%83%97')
+
     def test_equality(self):
         assert furl.Fragment() == furl.Fragment()
 
@@ -976,29 +996,57 @@ class TestFurl(unittest.TestCase):
         warnings.simplefilter("always")
 
     def _param(self, url, key, val):
-        # Note: urlparse.urlsplit() doesn't separate the query from the
-        # path for all schemes, only those schemes in the list
-        # urlparse.uses_query. So, as a result of using
-        # urlparse.urlsplit(), this little helper function only works
-        # when provided URLs whos schemes are also in
+        # urlsplit() only parses the query for schemes in urlparse.uses_query,
+        # so switch to 'http' (a scheme in urlparse.uses_query) for
+        # urlparse.urlsplit().
+        if '://' in url:
+            url = 'http://%s' % url.split('://', 1)[1]
+
+        # Note: urlparse.urlsplit() doesn't separate the query from the path
+        # for all schemes, only those schemes in the list urlparse.uses_query.
+        # So, as a result of using urlparse.urlsplit(), this little helper
+        # function only works when provided URLs whos schemes are also in
         # urlparse.uses_query.
         items = urllib.parse.parse_qsl(urllib.parse.urlsplit(url).query, True)
         return (key, val) in items
 
+    def test_idna(self):
+        decoded_host = u'ドメイン.テスト'
+        encoded_url = 'http://user:pass@xn--eckwd4c7c.xn--zckzah/'
+
+        f = furl.furl(encoded_url)
+        assert f.username == 'user' and f.password == 'pass'
+        assert f.host == decoded_host
+
+        f = furl.furl(encoded_url)
+        assert f.host == decoded_host
+
+        f = furl.furl('http://user:pass@pumps.ru/')
+        f.set(host=decoded_host)
+        assert f.url == encoded_url
+
+        f = furl.furl().set(host=u'ロリポップ')
+        assert f.url == 'xn--9ckxbq5co'
+
     def test_unicode(self):
-        paths = ['Восход', u'Восход']
+        paths = ['ロリポップ', u'ロリポップ']
         pairs = [('testö', 'testä'), (u'testö', u'testä')]
-        path_encoded = u'%D0%92%D0%BE%D1%81%D1%85%D0%BE%D0%B4'
         key_encoded, value_encoded = u'test%C3%B6', u'test%C3%A4'
+        path_encoded = u'%E3%83%AD%E3%83%AA%E3%83%9D%E3%83%83%E3%83%97'
 
         base_url = 'http://pumps.ru'
-        full_url = '%s/%s?%s=%s' % (
+        full_url_utf8_str = '%s/%s?%s=%s' % (
             base_url, paths[0], pairs[0][0], pairs[0][1])
+        full_url_unicode = u'%s/%s?%s=%s' % (
+            base_url, paths[1], pairs[1][0], pairs[1][1])
         full_url_encoded = '%s/%s?%s=%s' % (
             base_url, path_encoded, key_encoded, value_encoded)
 
+        f = furl.furl(full_url_utf8_str)
+        assert f.url == full_url_encoded
+
         # Accept unicode without raising an exception.
-        f = furl.furl(full_url)
+        f = furl.furl(full_url_unicode)
         assert f.url == full_url_encoded
 
         # Accept unicode paths.
@@ -1012,8 +1060,8 @@ class TestFurl(unittest.TestCase):
             f = furl.furl(base_url).set(path=path)
             f.args[key] = value
             assert f.args[key] == value  # Unicode values aren't modified.
-            assert not key in f.url
-            assert not value in f.url
+            assert key not in f.url
+            assert value not in f.url
             assert urllib.parse.quote_plus(furl.utf8(key)) in f.url
             assert urllib.parse.quote_plus(furl.utf8(value)) in f.url
             f.path.segments = [path]
@@ -1058,78 +1106,87 @@ class TestFurl(unittest.TestCase):
         f = furl.furl('mailto:')
         assert f.url == 'mailto:' and f.scheme == 'mailto'
 
+        # Ignore invalid schemes.
+        assert furl.furl('+invalid$scheme://lolsup').scheme is None
+        assert furl.furl('/api/test?url=http://a.com').scheme is None
+
     def test_username_and_password(self):
         # Empty usernames and passwords.
         for url in ['', 'http://www.pumps.com/']:
             f = furl.furl(url)
-            assert f.username is None and f.password is None
+            assert f.username is f.password is None
 
-        usernames = ['user', 'a-user_NAME$%^&09']
-        passwords = ['pass', 'a-PASS_word$%^&09']
         baseurl = 'http://www.google.com/'
+        usernames = ['', 'user', '@user', ' a-user_NAME$%^&09@:/']
+        passwords = ['', 'pass', ':pass', ' a-PASS_word$%^&09@:/']
 
         # Username only.
-        userurl = 'http://%s@www.google.com/'
         for username in usernames:
-            f = furl.furl(userurl % username)
+            encoded_username = urllib.parse.quote(username, safe='')
+            encoded_url = 'http://%s@www.google.com/' % encoded_username
+
+            f = furl.furl(encoded_url)
             assert f.username == username and f.password is None
 
             f = furl.furl(baseurl)
             f.username = username
             assert f.username == username and f.password is None
-            assert f.url == userurl % username
+            assert f.url == encoded_url
 
             f = furl.furl(baseurl)
             f.set(username=username)
             assert f.username == username and f.password is None
-            assert f.url == userurl % username
+            assert f.url == encoded_url
 
             f.remove(username=True)
-            assert f.username is None and f.password is None
-            assert f.url == baseurl
+            assert f.username is f.password is None and f.url == baseurl
 
         # Password only.
-        passurl = 'http://:%s@www.google.com/'
         for password in passwords:
-            f = furl.furl(passurl % password)
-            assert f.password == password and f.username is None
+            encoded_password = urllib.parse.quote(password, safe='')
+            encoded_url = 'http://:%s@www.google.com/' % encoded_password
+
+            f = furl.furl(encoded_url)
+            assert f.password == password and f.username == ''
 
             f = furl.furl(baseurl)
             f.password = password
-            assert f.password == password and f.username is None
-            assert f.url == passurl % password
+            assert f.password == password and not f.username
+            assert f.url == encoded_url
 
             f = furl.furl(baseurl)
             f.set(password=password)
-            assert f.password == password and f.username is None
-            assert f.url == passurl % password
+            assert f.password == password and not f.username
+            assert f.url == encoded_url
 
             f.remove(password=True)
-            assert not f.username and not f.password
-            assert f.url == baseurl
+            assert f.username is f.password is None and f.url == baseurl
 
         # Username and password.
-        userpassurl = 'http://%s:%s@www.google.com/'
         for username in usernames:
             for password in passwords:
-                f = furl.furl(userpassurl % (username, password))
+                encoded_username = urllib.parse.quote(username, safe='')
+                encoded_password = urllib.parse.quote(password, safe='')
+                encoded_url = 'http://%s:%s@www.google.com/' % (
+                    encoded_username, encoded_password)
+
+                f = furl.furl(encoded_url)
                 assert f.username == username and f.password == password
 
                 f = furl.furl(baseurl)
                 f.username = username
                 f.password = password
                 assert f.username == username and f.password == password
-                assert f.url == userpassurl % (username, password)
+                assert f.url == encoded_url
 
                 f = furl.furl(baseurl)
                 f.set(username=username, password=password)
                 assert f.username == username and f.password == password
-                assert f.url == userpassurl % (username, password)
+                assert f.url == encoded_url
 
                 f = furl.furl(baseurl)
                 f.remove(username=True, password=True)
-                assert not f.username and not f.password
-                assert f.url == baseurl
+                assert f.username is f.password is None and f.url == baseurl
 
         # Username and password in the network location string.
         f = furl.furl()
@@ -1283,7 +1340,7 @@ class TestFurl(unittest.TestCase):
         assert str(f.fragment) == ''
         assert f.url == ''
 
-        # Keep in mind that ';' is a query delimeter for both the URL
+        # Keep in mind that ';' is a query delimiter for both the URL
         # query and the fragment query, resulting in the str(path),
         # str(query), and str(fragment) values below.
         url = (
@@ -1347,6 +1404,14 @@ class TestFurl(unittest.TestCase):
         f = furl.furl('http://192.168.1.101')
         f = furl.furl('http://[2001:db8:85a3:8d3:1319:8a2e:370:7348]/')
 
+        # Host strings are always lowercase.
+        f = furl.furl('http://wWw.PuMpS.com')
+        assert f.host == 'www.pumps.com'
+        f.host = 'yEp.NoPe'
+        assert f.host == 'yep.nope'
+        f.set(host='FeE.fIe.FoE.fUm')
+        assert f.host == 'fee.fie.foe.fum'
+
         # Invalid IPv4 addresses shouldn't raise an exception because
         # urlparse.urlsplit() doesn't raise an exception on invalid IPv4
         # addresses.
@@ -1367,7 +1432,16 @@ class TestFurl(unittest.TestCase):
             with self.assertRaises(ValueError):
                 furl.furl('http://0:0:0:0:0:0:0:1]/')
 
-    def test_netlocs(self):
+        # Invalid host strings should raise ValueError.
+        invalid_hosts = ['.', '..', 'a..b', '.a.b', '.a.b.', '$', 'a$b']
+        for host in invalid_hosts:
+            with self.assertRaises(ValueError):
+                f = furl.furl('http://%s/' % host)
+        for host in invalid_hosts + ['a/b']:
+            with self.assertRaises(ValueError):
+                f = furl.furl('http://google.com/').set(host=host)
+
+    def test_netloc(self):
         f = furl.furl('http://pumps.com/')
         netloc = '1.2.3.4.5.6:999'
         f.netloc = netloc
@@ -1389,7 +1463,7 @@ class TestFurl(unittest.TestCase):
             with self.assertRaises(ValueError):
                 f.netloc = '0:0:0:0:0:0:0:1]'
 
-        # Invalid ports.
+        # Invalid ports should raise an exception.
         with self.assertRaises(ValueError):
             f.netloc = '[0:0:0:0:0:0:0:1]:alksdflasdfasdf'
         with self.assertRaises(ValueError):
@@ -1399,6 +1473,36 @@ class TestFurl(unittest.TestCase):
         assert f.host == '[0:0:0:0:0:0:0:1:1:1:1:1:1:1:1:9999999999999]'
         assert f.port == 888
 
+    def test_origin(self):
+        assert furl.furl().origin == '://'
+        assert furl.furl().set(host='slurp.ru').origin == '://slurp.ru'
+        assert furl.furl('http://pep.ru:83/yep').origin == 'http://pep.ru:83'
+        assert furl.furl().set(origin='pep://yep.ru').origin == 'pep://yep.ru'
+        f = furl.furl('http://user:pass@pumps.com/path?query#fragemtn')
+        assert f.origin == 'http://pumps.com'
+
+        f = furl.furl('none://ignored/lol?sup').set(origin='sup://yep.biz:99')
+        assert f.url == 'sup://yep.biz:99/lol?sup'
+
+        # Username and password are unaffected.
+        f = furl.furl('http://user:pass@slurp.com')
+        f.origin = 'ssh://horse-machine.de'
+        assert f.url == 'ssh://user:pass@horse-machine.de'
+
+        # Malformed IPv6 should raise an exception because
+        # urlparse.urlsplit() raises an exception in Python v2.7+.
+        if PYTHON_27PLUS:
+            with self.assertRaises(ValueError):
+                f.origin = '[0:0:0:0:0:0:0:1'
+            with self.assertRaises(ValueError):
+                f.origin = 'http://0:0:0:0:0:0:0:1]'
+
+        # Invalid ports should raise an exception.
+        with self.assertRaises(ValueError):
+            f.origin = '[0:0:0:0:0:0:0:1]:alksdflasdfasdf'
+        with self.assertRaises(ValueError):
+            f.origin = 'http://pump2pump.org:777777777777'
+
     def test_ports(self):
         # Default port values.
         assert furl.furl('http://www.pumps.com/').port == 80
@@ -1441,24 +1545,36 @@ class TestFurl(unittest.TestCase):
         assert f.port == 9999
         assert f.url == 'http://www.pumps.com:9999/'
 
-        self.assertRaises(f.set, port='asdf')
+        # The port is inferred from scheme changes, if possible, but
+        # only if the port is otherwise unset (self.port is None).
+        assert furl.furl('unknown://pump.com').set(scheme='http').port == 80
+        assert furl.furl('unknown://pump.com:99').set(scheme='http').port == 99
+        assert furl.furl('http://pump.com:99').set(scheme='unknown').port == 99
+
+        # Domains are always lowercase.
+        f = furl.furl('http://wWw.PuMpS.com:9999')
+        assert f.netloc == 'www.pumps.com:9999'
+        f.netloc = 'yEp.NoPe:9999'
+        assert f.netloc == 'yep.nope:9999'
+        f.set(netloc='FeE.fIe.FoE.fUm:9999')
+        assert f.netloc == 'fee.fie.foe.fum:9999'
 
     def test_add(self):
         f = furl.furl('http://pumps.com/')
 
-        assert f is f.add(args={'a': 'a', 'm': 'm&m'}, path='kwl jump',
+        assert f is f.add(args={'a': 'a', 'm': 'm&m'}, path='sp ace',
                           fragment_path='1', fragment_args={'f': 'frp'})
         assert self._param(f.url, 'a', 'a')
         assert self._param(f.url, 'm', 'm&m')
         assert str(f.fragment) == '1?f=frp'
-        assert str(f.path) == urllib.parse.urlsplit(f.url).path == '/kwl%20jump'
+        assert str(f.path) == urllib.parse.urlsplit(f.url).path == '/sp%20ace'
 
         assert f is f.add(path='dir', fragment_path='23', args={'b': 'b'},
                           fragment_args={'b': 'bewp'})
         assert self._param(f.url, 'a', 'a')
         assert self._param(f.url, 'm', 'm&m')
         assert self._param(f.url, 'b', 'b')
-        assert str(f.path) == '/kwl%20jump/dir'
+        assert str(f.path) == '/sp%20ace/dir'
         assert str(f.fragment) == '1/23?f=frp&b=bewp'
 
         # Supplying both <args> and <query_params> should raise a
@@ -1472,11 +1588,11 @@ class TestFurl(unittest.TestCase):
             assert params.index(('a', '1')) < params.index(('a', '2'))
 
     def test_set(self):
-        f = furl.furl('http://pumps.com/kwl%20jump/dir')
+        f = furl.furl('http://pumps.com/sp%20ace/dir')
         assert f is f.set(args={'no': 'nope'}, fragment='sup')
         assert 'a' not in f.args
         assert 'b' not in f.args
-        assert f.url == 'http://pumps.com/kwl%20jump/dir?no=nope#sup'
+        assert f.url == 'http://pumps.com/sp%20ace/dir?no=nope#sup'
 
         # No conflict warnings between <host>/<port> and <netloc>, or
         # <query> and <params>.
@@ -1520,17 +1636,29 @@ class TestFurl(unittest.TestCase):
         f = furl.furl('http://pumps.com')
         warnings.simplefilter("always")
 
-        # Host, port, and netloc overlap - host and port take
-        # precedence.
+        # Scheme, origin overlap. Scheme takes precedence.
+        with warnings.catch_warnings(record=True) as w1:
+            f.set(scheme='hi', origin='bye://sup.sup')
+            assert len(w1) == 1 and issubclass(w1[0].category, UserWarning)
+            assert f.scheme == 'hi'
+
+        # Netloc, origin, host and/or port. Host and port take precedence.
+        with warnings.catch_warnings(record=True) as w1:
+            f.set(netloc='dumps.com:99', origin='sup://pumps.com:88')
+            assert len(w1) == 1 and issubclass(w1[0].category, UserWarning)
         with warnings.catch_warnings(record=True) as w1:
             f.set(netloc='dumps.com:99', host='ohay.com')
             assert len(w1) == 1 and issubclass(w1[0].category, UserWarning)
-            f.host == 'ohay.com'
-            f.port == 99
+            assert f.host == 'ohay.com'
+            assert f.port == 99
         with warnings.catch_warnings(record=True) as w2:
             f.set(netloc='dumps.com:99', port=88)
             assert len(w2) == 1 and issubclass(w2[0].category, UserWarning)
-            f.port == 88
+            assert f.port == 88
+        with warnings.catch_warnings(record=True) as w2:
+            f.set(origin='http://dumps.com:99', port=88)
+            assert len(w2) == 1 and issubclass(w2[0].category, UserWarning)
+            assert f.port == 88
         with warnings.catch_warnings(record=True) as w3:
             f.set(netloc='dumps.com:99', host='ohay.com', port=88)
             assert len(w3) == 1 and issubclass(w3[0].category, UserWarning)
@@ -1618,33 +1746,37 @@ class TestFurl(unittest.TestCase):
                        'sup://www.pumps.org/brg/pap/mrf?a=b&c=d#frag?sup', ]
         run_tests = [
             # Join full URLs.
-            ('unknown://www.yahoo.com', 'unknown://www.yahoo.com'),
-            ('unknown://www.yahoo.com?one=two&three=four',
-             'unknown://www.yahoo.com?one=two&three=four'),
-            ('unknown://www.yahoo.com/new/url/?one=two#blrp',
-             'unknown://www.yahoo.com/new/url/?one=two#blrp'),
+            ('unknown://pepp.ru', 'unknown://pepp.ru'),
+            ('unknown://pepp.ru?one=two&three=four',
+             'unknown://pepp.ru?one=two&three=four'),
+            ('unknown://pepp.ru/new/url/?one=two#blrp',
+             'unknown://pepp.ru/new/url/?one=two#blrp'),
 
             # Absolute paths ('/foo').
-            ('/pump', 'unknown://www.yahoo.com/pump'),
-            ('/pump/2/dump', 'unknown://www.yahoo.com/pump/2/dump'),
-            ('/pump/2/dump/', 'unknown://www.yahoo.com/pump/2/dump/'),
+            ('/pump', 'unknown://pepp.ru/pump'),
+            ('/pump/2/dump', 'unknown://pepp.ru/pump/2/dump'),
+            ('/pump/2/dump/', 'unknown://pepp.ru/pump/2/dump/'),
 
             # Relative paths ('../foo').
-            ('./crit/', 'unknown://www.yahoo.com/pump/2/dump/crit/'),
-            ('.././../././././srp', 'unknown://www.yahoo.com/pump/2/srp'),
-            ('../././../nop', 'unknown://www.yahoo.com/nop'),
+            ('./crit/', 'unknown://pepp.ru/pump/2/dump/crit/'),
+            ('.././../././././srp', 'unknown://pepp.ru/pump/2/srp'),
+            ('../././../nop', 'unknown://pepp.ru/nop'),
 
             # Query included.
-            ('/erp/?one=two', 'unknown://www.yahoo.com/erp/?one=two'),
-            ('morp?three=four', 'unknown://www.yahoo.com/erp/morp?three=four'),
+            ('/erp/?one=two', 'unknown://pepp.ru/erp/?one=two'),
+            ('morp?three=four', 'unknown://pepp.ru/erp/morp?three=four'),
             ('/root/pumps?five=six',
-             'unknown://www.yahoo.com/root/pumps?five=six'),
+             'unknown://pepp.ru/root/pumps?five=six'),
 
             # Fragment included.
-            ('#sup', 'unknown://www.yahoo.com/root/pumps?five=six#sup'),
+            ('#sup', 'unknown://pepp.ru/root/pumps?five=six#sup'),
             ('/reset?one=two#yepYEP',
-             'unknown://www.yahoo.com/reset?one=two#yepYEP'),
-            ('./slurm#uwantpump?', 'unknown://www.yahoo.com/slurm#uwantpump?')
+             'unknown://pepp.ru/reset?one=two#yepYEP'),
+            ('./slurm#uwantpump?', 'unknown://pepp.ru/slurm#uwantpump?'),
+
+            # Unicode.
+            ('/?kødpålæg=4', 'unknown://pepp.ru/?k%C3%B8dp%C3%A5l%C3%A6g=4'),
+            (u'/?kødpålæg=4', 'unknown://pepp.ru/?k%C3%B8dp%C3%A5l%C3%A6g=4'),
         ]
 
         for test in empty_tests:
@@ -1661,18 +1793,34 @@ class TestFurl(unittest.TestCase):
             tojoin = furl.furl(join)
             assert f is f.join(tojoin) and f.url == result
 
+        # Join multiple URLs.
+        f = furl.furl('')
+        f.join('path', 'tcp://blorp.biz', 'http://pepp.ru/', 'a/b/c',
+               '#uwantpump?')
+        assert f.url == 'http://pepp.ru/a/b/c#uwantpump?'
+
+    def test_tostr(self):
+        f = furl.furl('http://blast.off/?a+b=c+d&two%20tap=cat%20nap%24')
+        assert f.tostr() == f.url
+        assert (f.tostr(query_delimiter=';') ==
+                'http://blast.off/?a+b=c+d;two+tap=cat+nap$')
+        assert (f.tostr(query_quote_plus=False) ==
+                'http://blast.off/?a%20b=c%20d&two%20tap=cat%20nap$')
+        assert (f.tostr(query_delimiter=';', query_quote_plus=False) ==
+                'http://blast.off/?a%20b=c%20d;two%20tap=cat%20nap$')
+
     def test_equality(self):
         assert furl.furl() is not furl.furl() and furl.furl() == furl.furl()
 
-        assert furl.furl() != None
+        assert furl.furl() is not None
 
         url = 'https://www.yahoo.co.uk/one/two/three?a=a&b=b&m=m%26m#fragment'
-        assert furl.furl(url) != url # No furl to string comparisons (for now).
+        assert furl.furl(url) != url  # No furl to string comparisons.
         assert furl.furl(url) == furl.furl(url)
         assert furl.furl(url).remove(path=True) != furl.furl(url)
 
     def test_urlsplit(self):
-        # Without any delimeters like '://' or '/', the input should be
+        # Without any delimiters like '://' or '/', the input should be
         # treated as a path.
         urls = ['sup', '127.0.0.1', 'www.google.com', '192.168.1.1:8000']
         for url in urls:
@@ -1869,7 +2017,17 @@ class TestFurl(unittest.TestCase):
         for port in invalids:
             assert not furl.is_valid_port(port)
 
-    def test_is_valid_encoded_path_segment(segment):
+    def test_is_valid_scheme(self):
+        valids = ['a', 'ab', 'a-b', 'a.b', 'a+b', 'a----b', 'a123', 'a-b123',
+                  'a+b.1-+']
+        invalids = ['1', '12', '12+', '-', '.', '+', '1a', '+a', '.b.']
+
+        for valid in valids:
+            assert furl.is_valid_scheme(valid)
+        for invalid in invalids:
+            assert not furl.is_valid_scheme(invalid)
+
+    def test_is_valid_encoded_path_segment(self):
         valids = [('abcdefghijklmnopqrstuvwxyz'
                    'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
                    '0123456789' '-._~' ":@!$&'()*+,;="),
@@ -1881,7 +2039,7 @@ class TestFurl(unittest.TestCase):
         for invalid in invalids:
             assert not furl.is_valid_encoded_path_segment(invalid)
 
-    def test_is_valid_encoded_query_key(key):
+    def test_is_valid_encoded_query_key(self):
         valids = [('abcdefghijklmnopqrstuvwxyz'
                    'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
                    '0123456789' '-._~' ":@!$&'()*+,;" '/?'),
@@ -1893,7 +2051,7 @@ class TestFurl(unittest.TestCase):
         for invalid in invalids:
             assert not furl.is_valid_encoded_query_key(invalid)
 
-    def test_is_valid_encoded_query_value(value):
+    def test_is_valid_encoded_query_value(self):
         valids = [('abcdefghijklmnopqrstuvwxyz'
                    'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
                    '0123456789' '-._~' ":@!$&'()*+,;" '/?='),
